from __future__ import annotations
from typing import TYPE_CHECKING, List
import os
import shlex
import mama.util
import mama.utils.sub_process as proc
from mama.utils.system import console

if TYPE_CHECKING:
    from mama.build_target import BuildTarget

######################################################################################

class BuildProduct:
    """ Represents the build product of a project that should be deployed """
    def __init__(self, built_path:str, deploy_path:str=None, strip=None, is_dir=False):
        """
            - built_path: where the built file exists.
                          Supported project variables {{installed}}, {{source}}, {{build}}
            - deploy_path where the file should be deployed, can be None if no deploy needed
            - strip: whether to strip the binary when deploying
            - is_dir: whether the built_path is a directory
        """
        self.built_path = built_path
        self.deploy_path = deploy_path
        # if strip option is specified, then use it,
        # otherwise default to strip=True for files and strip=False for dirs
        self.strip = strip if strip != None else not is_dir
        self.is_dir = is_dir

######################################################################################

class GnuProject:
    """ 
    Helper utility class for compiling GNU projects.
    This will enable downloading, unzipping and building GNU projects.

    Example usage:
    ```
    gmp = GnuProject(target, 'gmp', '6.2.1', 'https://gmplib.org/download/gmp/{{project}}.tar.xz', 'lib/libgmp.a')
    gmp.build()
    """
    def __init__(self, target:BuildTarget, name:str, version:str,
                 url:str='',
                 git:str='',
                 build_products=[],
                 autogen=False,
                 configure='configure'):
        """
        - target: mama.BuildTarget building this GnuProject and collecting the build products
        - name: name of the project, eg 'gmp'
        - version: version of the project, eg '6.2.1'
        - url: url to download the project, eg 'https://gmplib.org/download/gmp/{{project}}.tar.xz'
        - git: git to clone the project from
        - build_products: the final products to build, eg [BuildProduct('{{installed}}/lib/libgmp.a', 'mypath/libgmp.a')].
                          Supported project variables {{installed}}, {{source}}, {{build}}
        - autogen: whether to use ./autogen.sh before running ./configure
        - configure: the configuration command, by default 'configure' but can be 'make config' etc
        """
        self.target = target
        self.name = name
        self.version = version
        self.name_with_version = f'{name}-{version}'
        self.url = url
        self.git = git
        self.autogen = autogen
        if isinstance(build_products, list):
            self.build_products = build_products
        elif isinstance(build_products, BuildProduct):
            self.build_products = [build_products]
        else:
            raise RuntimeError('build_product must be a BuildProduct or a list of BuildProducts')
        self.install_dir_suffix = '-built'
        self.make_opts = '' # default options for make
        self.host = ''
        if self.target.config.mips:
            self.host = 'mipsel-linux-gnu'
        elif self.target.config.oclea:
            self.host = 'aarch64-oclea-linux'
        # the configure command, by default it's 'configure'
        # however using something other than 'configure' will completely override it
        self.configure_command = configure
        # extra environment variables to set when running the project
        self.extra_env = {}


    def source_dir(self, subpath=''):
        """ Where the project is extracted to, eg project/mips/gdb-13.2 """
        if subpath:
            return self.target.build_dir(self.name_with_version + '/' + subpath)
        return self.target.build_dir(self.name_with_version)


    def install_dir(self, subpath=''):
        """ Where the project is installed to, eg project/mips/gdb-built """
        if subpath:
            return self.target.build_dir(self.name + self.install_dir_suffix + '/' + subpath)
        return self.target.build_dir(self.name + self.install_dir_suffix)


    def add_build_product(self, product:BuildProduct):
        self.build_products.append(product)


    def get_parsed_path(self, path:str):
        """ Parses the path with the project variables: {{installed}}, {{source}}, {{build}} """
        if '{{installed}}' in path:
            path = path.replace('{{installed}}', self.install_dir())
        if '{{source}}' in path:
            path = path.replace('{{source}}', self.source_dir())
        if '{{build}}' in path:
            path = path.replace('{{build}}', self.target.build_dir())
        return path


    def should_build(self):
        """ Returns True if any deployed files or any built files are missing """
        if self.has_deployables() and not self.should_deploy():
            return False
        for p in self.build_products:
            if not os.path.exists(self.get_parsed_path(p.built_path)):
                return True
        return False


    def should_deploy(self):
        """ Returns True if any of the deployed files are missing """
        for f in self.build_products:
            if f.deploy_path and not os.path.exists(f.deploy_path):
                return True
        return False


    def has_deployables(self):
        """ Returns True if any of the build products have a deploy path """
        for f in self.build_products:
            if f.deploy_path:
                return True
        return False


    def deploy_all_products(self, force=False):
        """ Deploys all built products if needed """
        if force or self.should_deploy():
            for p in self.build_products:
                if p.deploy_path:
                    if p.is_dir:
                        self.deploy_dir(p.built_path, p.deploy_path, strip=p.strip)
                    else:
                        self.deploy(p.built_path, p.deploy_path, strip=p.strip)
            return True
        return False


    def get_makefile(self):
        """ Gets the Makefile, which is a build step dependency """
        return f'{self.source_dir()}/Makefile'


    def get_configure_file(self):
        """ Gets the file which performs configuration, and is thus a configuration step dependency """
        args = shlex.split(self.configure_command)
        config_cmd = args[0].strip()
        if config_cmd == 'make':  # if config_cmd is 'make', then look for a Makefile
            return self.get_makefile()
        return f'{self.source_dir()}/{config_cmd}'


    def checkout_code(self):
        """
        Checks out the code archive, either by downloading and extracting the zip archive, 
        or cloning the project from git repository.
        """
        source = self.source_dir()
        configure_file = self.get_configure_file()
        autogen_file = f'{self.source_dir()}/autogen.sh' if self.autogen else ''
        if autogen_file:
            if os.path.exists(autogen_file):
                return # nothing to do
        elif os.path.exists(configure_file):
            return # nothing to do

        build_root = self.target.build_dir()
        if self.git:
            console(f'>>> Cloning {source} from {self.git}', color='green')
            if os.system(f'git clone {self.git} {source}') != 0:
                raise Exception(f'Failed to clone {self.git} to {source}')
        else:
            url = self.url.replace('{{project}}', self.name_with_version)
            try:
                local_file = mama.util.download_file(url, local_dir=build_root)
            except Exception as e:
                raise Exception(f'Failed to download {url}: {e}')

            console(f'>>> Extracting to {source}', color='green')
            os.makedirs(source, exist_ok=True)
            if local_file.endswith('.tar.xz') \
                or local_file.endswith('.tar.gz') \
                or local_file.endswith('.tar.gz2') \
                or local_file.endswith('.tar.bz2'):
                proc.execute_echo(build_root, f'tar -xf {local_file} -C {source} --strip-components=1')
            elif local_file.endswith('.zip'):
                mama.util.unzip(local_file, source)
            else:
                console(f'>>> ERROR: Unknown archive type: {local_file}', color='red')

        # final check if the configure file exists
        if autogen_file:
            if not os.path.exists(autogen_file):
                raise Exception(f'Checkout failed, no autogen file at: {autogen_file}')
        else:
            if not os.path.exists(configure_file):
                raise Exception(f'Checkout failed, no configure file at: {configure_file}')


    def configure_env(self):
        if self.target.oclea:
            self.target.oclea.get_gnu_build_env(self.extra_env)
        else:
            # GNU projects need to be configured with the CC, CXX and AR environment variables set
            cc_prefix = self.target.get_cc_prefix()
            if cc_prefix:
                os.environ['CC'] = cc_prefix + 'gcc'
                os.environ['CXX'] = cc_prefix + 'g++'
                os.environ['AR'] = cc_prefix + 'ar'
                os.environ['LD'] = cc_prefix + 'ld'
                os.environ['READELF'] = cc_prefix + 'readelf'
                os.environ['STRIP'] = cc_prefix + 'strip'
                os.environ['RANLIB'] = cc_prefix + 'ranlib'



    def run(self, command):
        """ Runs a command in the project directory, eg 'make specialsetup' """
        env = None
        if self.extra_env: # copy env and add extra env vars
            env = os.environ.copy()
            for k, v in self.extra_env.items():
                env[k] = v
        self.target.run_program(self.source_dir(), command, env=env)


    def configure(self, options='', autogen_opts='', prefix=''):
        """
        Only configures the project for building by generating the Makefile
            - options: additional options to pass to the configure script
        """
        console(f'>>> Configuring {self.name}', color='green')
        self.configure_env()

        if self.autogen:
            self.run(f'./autogen.sh {autogen_opts}')

        if self.configure_command != 'configure':
            # user overrides with custom configurator, such as `make config` etc
            configure = self.configure_command
        else:
            args = ''
            if self.host:
                args += f' --host={self.host}'
            if not prefix:
                prefix = f'--prefix {self.install_dir()}'
            if not self.autogen:
                guess_machine = f'{self.source_dir()}/config.guess'
                if os.path.exists(guess_machine):
                    os.chmod(guess_machine, 0o755) # make sure it's executable
                    args += f' --build={proc.execute_piped(guess_machine)}'
            configure = f'./configure {args} {prefix}'

        self.run(f'{configure} {options}')
        console(f'>>> Configured {self.name}', color='green')


    def _get_make_opts(self, opts, multithreaded=False):
        jobs = f'-j {self.target.config.jobs}' if multithreaded else ''
        all_opts = ''
        for o in [self.make_opts, opts, jobs]:
            if all_opts: all_opts += ' '
            all_opts += o
        return all_opts


    def make(self, opts='', multithreaded=False):
        """
        Only makes the project
            - opts: extra options for make
            - multithreaded: if true, will use the -j option to build with multiple threads
        """
        make_opts = self._get_make_opts(opts, multithreaded)
        console(f'>>> Make {self.name} {make_opts}', color='green')
        self.configure_env()
        self.run(f'make {make_opts}')
        console(f'>>> Made {self.name} {make_opts}', color='green')


    def install(self, no_prefix=False):
        """ Only installs the project """
        console(f'>>> Installing {self.name}', color='green')
        self.configure_env()
        prefix = '' if no_prefix else f'PREFIX={self.install_dir()}'
        all_opts = self._get_make_opts(prefix, multithreaded=False)
        self.run(f'make {all_opts} install')
        console(f'>>> Installed {self.name}', color='green')
        return self.install_dir()


    def build(self, options='', make_opts='', prefix='',
              multithreaded=False, install=True):
        """
        Downloads, Unzips, Configures, Makes and Installs the project 
            - options: additional options to pass to the configure script
            - multithreaded: if true, will use the -j option to build with multiple threads
        """
        project_dir = self.source_dir()
        autoconf_makefile = f'{project_dir}/Makefile' if self.configure_command == 'configure' else None
        try:
            console(f'>>>>>> BUILD {self.name} <<<<<<', color='green')
            self.checkout_code()

            if not autoconf_makefile: # if not autoconf then always run the configure step
                self.configure(options=options, prefix=prefix)
            elif not os.path.exists(autoconf_makefile):
                self.configure(options=options, prefix=prefix)

            self.make(opts=make_opts, multithreaded=multithreaded)
            if install:
                self.install()
        except:
            console(f'>>> ERROR: Failed to build {self.name}', color='red')
            # with autoconf projects, delete the makefile so that it will be regenerated
            #if autoconf_makefile and os.path.exists(autoconf_makefile):
            #    os.remove(autoconf_makefile)
            raise


    def strip(self, src_path, dest_path=None):
        prefix = self.target.get_cc_prefix()
        striptool = prefix + 'strip' if prefix else 'strip'
        out = f'-o {dest_path}' if dest_path else ''
        if os.system(f'{striptool} {src_path} {out}') != 0:
            raise Exception(f'Failed to strip {src_path}')


    def copy_file_or_link(self, src_file, dst_file):
        """ Copies a file or symlink preserving their attributes and relative symlinks """
        if os.path.islink(src_file):
            link = os.readlink(src_file)
            #console(f'link: {dst_file} -> {link}', color='yellow')
            os.remove(dst_file)
            os.symlink(link, dst_file)
        else:
            mama.util.copy_if_needed(src_file, dst_file)


    def can_strip(self, filepath):
        if not os.path.isfile(filepath):
            return False
        ext = os.path.splitext(filepath)[1]
        return ext == None or ext == ''


    def deploy(self, src_path:str, dest_path=None, strip=False):
        """ 
        Deploys the src_path to dest_path, 
        optionally stripping the binary in the process.
        """
        if not src_path:
            raise RuntimeError('src_path must be specified')

        src_path = self.get_parsed_path(src_path)
        dest_path = self.get_parsed_path(dest_path) if dest_path else src_path
        dest_dir = os.path.dirname(dest_path)
        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir, exist_ok=True)

        if strip and self.can_strip(src_path):
            self.strip(src_path, dest_path)
        else:
            self.copy_file_or_link(src_path, dest_path)
        if not os.path.exists(dest_path):
            raise Exception(f'Failed to deploy {src_path} to {dest_path}')
        console(f'>>> Deployed {src_path} to {dest_path}', color='green')


    def deploy_dir(self, src_dir, dest_dir, strip=False):
        """ Deploys all built products from src_dir to dest_dir and optionally strips them """
        src_dir = self.get_parsed_path(src_dir)
        if not os.path.exists(src_dir):
            raise Exception(f'Failed to deploy from {src_dir}, directory does not exist')

        dest_dir = self.get_parsed_path(dest_dir)
        root = os.path.dirname(src_dir)
        count = 0
        for fulldir, _, files in os.walk(src_dir):
            reldir = fulldir[len(root):].lstrip('\\/')
            for file in files:
                if reldir:
                    dst_folder = os.path.join(dest_dir, reldir)
                else:
                    dst_folder = dest_dir
                os.makedirs(dst_folder, exist_ok=True)
                src_file = os.path.join(fulldir, file)
                dst_file = os.path.join(dst_folder, file)
                if strip and self.can_strip(src_file):
                    self.strip(src_file, dest_path=dst_file)
                else:
                    self.copy_file_or_link(src_file, dst_file)
                count += 1

        if count > 0:
            console(f'>>> Deployed {src_dir} to {dest_dir}', color='green')
        else:
            console(f'>>> No files to deploy from {src_dir}', color='yellow')


######################################################################################




from __future__ import annotations
from typing import List, TYPE_CHECKING
import os

from .types.git import Git
from .types.local_source import LocalSource
from .types.artifactory_pkg import ArtifactoryPkg
from .types.dep_source import DepSource
from .types.asset import Asset

from .util import normalized_path, normalized_join, read_lines_from \
                , write_text_to, console, copy_if_needed

import mama.package as package

if TYPE_CHECKING:
    from .build_target import BuildTarget
    from .build_dependency import BuildDependency


def _gather_dependencies(target:BuildTarget) -> List[BuildDependency]:
    dependecies = []
    for child in target.children():
        dependecies.append(child)
    return dependecies


def _results_contain(results, contains_value):
    for target,value in results:
        if value == contains_value:
            return True
    return False


def _gather(target:BuildTarget, recurse, results:list, get_candidates):
    for value in get_candidates(target):
        if not _results_contain(results, value):
            results.append((target,value))
    if recurse:
        for child in target.children():
            _gather(child.target, True, results, get_candidates)
    return results


def _gather_includes(target:BuildTarget, recurse):
    includes = []
    return _gather(target, recurse, includes, lambda t: t.exported_includes)


def _gather_libs(target:BuildTarget, recurse):
    # gather all libs from the root target
    libs = [(target,l) for l in target.exported_libs]

    # and for children, only gather dynamic libs if recurse is set
    if recurse:
        def get_dylibs(t:BuildTarget):
            for l in t.exported_libs:
                if package.is_a_dynamic_library(l): yield l
        for child in target.children():
            _gather(child.target, recurse, libs, get_dylibs)
    return libs


def _gather_syslibs(target:BuildTarget, recurse):
    syslibs = []
    return _gather(target, recurse, syslibs, lambda t: t.exported_syslibs)


def _gather_assets(target:BuildTarget, recurse):
    assets = []
    return _gather(target, recurse, assets, lambda t: t.exported_assets)


def _append_includes(target:BuildTarget, package_full_path, detail_echo, descr, includes):
    if not includes:
        return # nothing to do
    config = target.config
    includes_root = package_full_path + '/include'
    # TODO: should we include .cpp files for easier debugging?
    includes_filter = ['.h','.hpp','.hxx','.hh','.c','.cpp','.cxx']

    # set the default include
    descr.append(f'I include')

    def append(relpath):
        src_path = abs_include
        dst_dir = includes_root

        # matches default include?
        if relpath == 'include' or relpath == 'include/':
            if detail_echo: console(f'    I ({inctarget.name+")": <16}  include')
            dst_dir = normalized_path(includes_root + '/../')
        else:
            if detail_echo: console(f'    I ({inctarget.name+")": <16}  include/{relpath}')
            descr.append(f'I include/{relpath}')

        src_dir = os.path.dirname(src_path)
        if src_dir != dst_dir:
            if config.verbose: console(f'    copy {src_path}\n      -> {dst_dir}')
            copy_if_needed(src_path, dst_dir, includes_filter)

    relincludes = []  # TODO: what was the point of this again?
    for inctarget, abs_include in includes:
        relpath = os.path.basename(abs_include)
        if not relpath in relincludes:
            relincludes.append(relpath)
            append(relpath)


def papa_deploy_to(target:BuildTarget, package_full_path:str,
                   r_includes:bool, r_dylibs:bool, 
                   r_syslibs:bool, r_assets:bool):
    config = target.config
    detail_echo = config.print and target.is_current_target() and (not config.test)
    if detail_echo: console(f'  - PAPA Deploy {package_full_path}')

    dependencies = _gather_dependencies(target)

    if not os.path.exists(package_full_path): # check to avoid Access Denied errors
        os.makedirs(package_full_path, exist_ok=True)

    # set up project and dependencies
    descr = [ f'P {target.name}' ]
    for d in dependencies:
        if detail_echo: console(f'    D {d.dep_source}')
        descr.append(f'D {d.dep_source.get_papa_string()}')

    includes = _gather_includes(target, r_includes)
    _append_includes(target, package_full_path, detail_echo, descr, includes)

    build_dir = target.build_dir()
    source_dir = target.source_dir()

    libs = _gather_libs(target, r_dylibs)
    for libtarget, lib in libs:
        if   lib.startswith(build_dir):  relpath = os.path.relpath(lib, build_dir)
        elif lib.startswith(source_dir): relpath = os.path.relpath(lib, source_dir)
        else: relpath = lib
        descr.append(f'L {relpath}')
        outpath = normalized_join(package_full_path, relpath)
        os.makedirs(os.path.dirname(outpath), exist_ok=True)
        #outpath = package_full_path
        if detail_echo: console(f'    L ({libtarget.name+")": <16}  {relpath}')
        if lib != outpath:
            if config.verbose: console(f'    copy {lib}\n      -> {outpath}')
            copy_if_needed(lib, outpath)

    syslibs = _gather_syslibs(target, r_syslibs)
    for systarget, syslib in syslibs:
        syslib_basename = package.get_lib_basename(syslib)
        descr.append(f'S {syslib_basename}')
        if detail_echo: console(f'    S ({systarget.name+")": <16}  {syslib_basename}')

    assets = _gather_assets(target, r_assets)
    for asstarget, asset in assets:
        descr.append(f'A {asset.outpath}')
        if detail_echo: console(f'    A ({asstarget.name+")": <16}  {asset.outpath}')
        outpath = normalized_join(package_full_path, asset.outpath)
        if asset.srcpath != outpath:
            folder = os.path.dirname(outpath)
            if not os.path.exists(folder):
                os.makedirs(folder, exist_ok=True)
            copy_if_needed(asset.srcpath, outpath)

    write_text_to(os.path.join(package_full_path, 'papa.txt'), '\n'.join(descr))

    # write summary
    if config.print:
        console(f'  PAPA Deployed: {len(includes)} includes, {len(libs)} libs, {len(syslibs)} syslibs, {len(assets)} assets')


def make_dep_source(s:str) -> DepSource:
    if s.startswith('git '): return Git.from_papa_string(s[4:])
    if s.startswith('pkg '): return ArtifactoryPkg.from_papa_string(s[4:])
    if s.startswith('src '): return LocalSource.from_papa_string(s[4:])
    raise RuntimeError(f'Unrecognized dependency source: {s}')


class PapaFileInfo:
    def __init__(self, papa_file:str):
        if not os.path.exists(papa_file):
            raise FileNotFoundError(f'Package file not found: {papa_file}')
        self.papa_file = papa_file
        self.papa_dir = os.path.dirname(papa_file)

        self.project_name = None
        self.dependencies = []
        self.includes = []
        self.libs = []
        self.syslibs = []
        self.assets = []

        def append_to(to:list, line):
            to.append(normalized_join(self.papa_dir, line[2:].strip()))

        for line in read_lines_from(self.papa_file):
            if   line.startswith('P '): self.project_name = line[2:].strip()
            elif line.startswith('D '): self.dependencies.append(make_dep_source(line[2:].strip()))
            elif line.startswith('I '): append_to(self.includes, line)
            elif line.startswith('L '): append_to(self.libs, line)
            elif line.startswith('S '): append_to(self.syslibs, line)
            elif line.startswith('A '):
                relpath = line[2:].strip()
                fullpath = normalized_join(self.papa_dir, relpath)
                self.assets.append(Asset(relpath, fullpath, None))




import os
from typing import Callable
from mama.utils.system import System, console
from mama.util import path_join, read_lines_from

class Mips:
    def __init__(self, config):
        self.name = 'mips'
        self.toolchain_major = 1
        self.toolchain_minor = 0
        self.config = config
        self.toolchain_dir = None
        self.toolchain_file = None
        self.supported_arches = ['mips', 'mipsel', 'mips64', 'mips64el']
        self.mips_arch = 'mipsel' # prefer little endian mips
        self.gcc_prefix = '' # prefix to gcc binary
        self.libs_path = '' # toolchain lib/ path
        self.include_path = '' # toolchain include/ path


    # returns the current path prefix where the compiler can be found
    # example: "/opt/mipsel-openwrt-linux/bin/mipsel-openwrt-linux-"
    def compiler_prefix(self):
        if not self.gcc_prefix: self.init_default()
        return self.gcc_prefix


    # forced includes that should be added to compiler flags as -I paths
    def includes(self):
        return []


    def init_default(self):
        if not self.gcc_prefix:
            self.init_toolchain(self.mips_arch)


    def init_toolchain(self, arch, toolchain_dir=None, toolchain_file=None):
        if arch not in self.supported_arches:
            raise RuntimeError(f'Unsupported MIPS arch: {arch}')
        if toolchain_file and not os.path.exists(toolchain_file):
            raise FileNotFoundError(f'Toolchain file not found: {toolchain_file}')
        if toolchain_dir and not os.path.exists(toolchain_dir):
            raise FileNotFoundError(f'Toolchain directory not found: {toolchain_dir}')
        if not System.linux:
            raise RuntimeError('MIPS only supported on Linux')

        # check if we have already initialized the toolchain
        if self.gcc_prefix and self.mips_arch == arch \
            and self.toolchain_file == toolchain_file \
            and self.toolchain_dir == toolchain_dir:
            return

        # direct system installed MIPS toolchain
        self.mips_arch = arch
        self.toolchain_file = toolchain_file # additional toolchain to specify sysroot details

        # if a toolchain dir is provided, it should have a bin/ subdir with the compiler
        if toolchain_dir:
            self.toolchain_dir = toolchain_dir
            if os.path.exists(f'{toolchain_dir}/bin/{arch}-linux-gnu-gcc'):
                gcc_prefix = f'{toolchain_dir}/bin/{arch}-linux-gnu-'
                libs_path = f'{toolchain_dir}/lib'
                include_path = f'{toolchain_dir}/include'
                self._set_mips_toolchain_dir(gcc_prefix, libs_path, include_path)
                return # success

        # might also be at `/usr/mipsel-linux-gnu`

        # check for system installed one as fallback
        if os.path.exists(f'/usr/bin/{arch}-linux-gnu-gcc'):
            gcc_prefix = f'/usr/bin/{arch}-linux-gnu-'
            libs_path = f'/usr/{arch}-linux-gnu/lib'
            include_path = f'/usr/{arch}-linux-gnu/include'
            self._set_mips_toolchain_dir(gcc_prefix, libs_path, include_path)
            return # success

        raise EnvironmentError('No MIPS toolchain compilers detected, '+
                               f'try "sudo apt-get install g++-{arch}-linux-gnu"')

    def _set_mips_toolchain_dir(self, gcc_prefix, libs_path, include_path):
        self.gcc_prefix = gcc_prefix
        self.libs_path = libs_path if os.path.exists(libs_path) else ''
        self.include_path = include_path if os.path.exists(include_path) else ''

        # attempt to detect toolchain version from gcc linux/version.h
        try:
            version_file = path_join(self.include_path, 'linux/version.h')
            if os.path.exists(version_file):
                for line in read_lines_from(version_file):
                    if line.startswith('#define LINUX_VERSION_MAJOR'):
                        self.toolchain_major = int(line.split()[2])
                    elif line.startswith('#define LINUX_VERSION_PATCHLEVEL'):
                        self.toolchain_minor = int(line.split()[2])
        except:
            pass
        if self.config.print:
            console(f'Found MIPS tools: {self.gcc_prefix}gcc  linux-v{self.toolchain_major}.{self.toolchain_minor}')
            if self.libs_path:
                console(f'  MIPS syslibs: {self.libs_path}')


    def get_cxx_flags(self, add_flag: Callable[[str,str], None]):
        add_flag('-DMIPS', '1')
        if self.libs_path:
            add_flag(f'-L {self.libs_path}')
        for path in self.includes():
            add_flag(f'-I {path}')

    def get_cmake_build_opts(self) -> list:
        if self.toolchain_file:
            if self.config.print:
                console(f'MIPS Toolchain: {self.toolchain_file}')
            return [
                'MIPS=TRUE',
                f'CMAKE_TOOLCHAIN_FILE="{self.toolchain_file}"'
            ]
        opt = [
            'MIPS=TRUE',
            'CMAKE_SYSTEM_NAME=Linux',
            'CMAKE_SYSTEM_VERSION=1',
            f'CMAKE_SYSTEM_PROCESSOR={self.mips_arch}',
            'CMAKE_FIND_ROOT_PATH_MODE_PROGRAM=ONLY', # Search for compiler tools
            'CMAKE_FIND_ROOT_PATH_MODE_LIBRARY=ONLY', # Search for libraries and headers in the target directories only
            'CMAKE_FIND_ROOT_PATH_MODE_INCLUDE=ONLY',
        ]
        return opt
